@@ -1,6 +1,6 @@
-//Copyright 2003-2005 Arthur van Hoff Rick Blair
-//Licensed under Apache License version 2.0
-//Original license LGPL
+// Copyright 2003-2005 Arthur van Hoff Rick Blair
+// Licensed under Apache License version 2.0
+// Original license LGPL
 
 package javax.jmdns.impl;
 
@@ -9,6 +9,7 @@ import java.util.ArrayList;
 import java.util.Collection;
 import java.util.Collections;
 import java.util.HashSet;
+import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
 import java.util.Set;
@@ -19,147 +20,123 @@ import javax.jmdns.impl.constants.DNSRecordType;
 /**
  * A table of DNS entries. This is a map table which can handle multiple entries with the same name.
  * <p/>
- * Storing multiple entries with the same name is implemented using a linked list. This is hidden from the user and can
- * change in later implementation.
+ * Storing multiple entries with the same name is implemented using a linked list. This is hidden from the user and can change in later implementation.
  * <p/>
  * Here's how to iterate over all entries:
- *
+ * 
  * <pre>
  *       for (Iterator i=dnscache.allValues().iterator(); i.hasNext(); ) {
  *             DNSEntry entry = i.next();
  *             ...do something with entry...
  *       }
  * </pre>
- *
  * <p/>
  * And here's how to iterate over all entries having a given name:
- *
+ * 
  * <pre>
  *       for (Iterator i=dnscache.getDNSEntryList(name).iterator(); i.hasNext(); ) {
  *             DNSEntry entry = i.next();
  *           ...do something with entry...
  *       }
  * </pre>
- *
- * @version %I%, %G%
+ * 
  * @author Arthur van Hoff, Werner Randelshofer, Rick Blair, Pierre Frisch
  */
-public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
-{
+public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>> {
+
     // private static Logger logger = Logger.getLogger(DNSCache.class.getName());
 
-    // Implementation note:
-    // We might completely hide the existence of CacheNode's in a future version
-    // of DNSCache. But this will require to implement two (inner) classes for
-    // the iterators that will be returned by method <code>iterator()</code> and
-    // method <code>find(name)</code>.
-    // Since DNSCache is not a public class, it does not seem worth the effort
-    // to clean its API up that much.
-
-    // [PJYF Oct 15 2004] This should implements Collections that would be a much cleaner implementation
-    private transient Set<Map.Entry<String, List<? extends DNSEntry>>>    _entrySet   = null;
+    private transient Set<Map.Entry<String, List<? extends DNSEntry>>> _entrySet  = null;
 
     /**
      *
      */
-    public static final DNSCache                                          EmptyCache  = new _EmptyCache();
+    public static final DNSCache                                       EmptyCache = new _EmptyCache();
 
-    static class _EmptyCache extends DNSCache {
+    static final class _EmptyCache extends DNSCache {
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.AbstractMap#size()
+        /**
+         * {@inheritDoc}
          */
         @Override
         public int size() {
             return 0;
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.AbstractMap#isEmpty()
+        /**
+         * {@inheritDoc}
          */
         @Override
         public boolean isEmpty() {
             return true;
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.AbstractMap#containsKey(java.lang.Object)
+        /**
+         * {@inheritDoc}
          */
         @Override
         public boolean containsKey(Object key) {
             return false;
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.AbstractMap#containsValue(java.lang.Object)
+        /**
+         * {@inheritDoc}
          */
         @Override
         public boolean containsValue(Object value) {
             return false;
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.AbstractMap#get(java.lang.Object)
+        /**
+         * {@inheritDoc}
          */
         @Override
         public List<DNSEntry> get(Object key) {
             return null;
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.AbstractMap#keySet()
+        /**
+         * {@inheritDoc}
          */
         @Override
         public Set<String> keySet() {
             return Collections.emptySet();
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.AbstractMap#values()
+        /**
+         * {@inheritDoc}
          */
         @Override
         public Collection<List<? extends DNSEntry>> values() {
             return Collections.emptySet();
         }
 
-        /*
-         * (non-Javadoc)
-         * @see com.webobjects.discoveryservices.DNSCache#entrySet()
+        /**
+         * {@inheritDoc}
          */
         @Override
         public Set<Map.Entry<String, List<? extends DNSEntry>>> entrySet() {
             return Collections.emptySet();
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.AbstractMap#equals(java.lang.Object)
+        /**
+         * {@inheritDoc}
          */
-        @SuppressWarnings("unchecked")
         @Override
         public boolean equals(Object o) {
-            return (o instanceof Map) && ((Map) o).size() == 0;
+            return (o instanceof Map) && ((Map<?, ?>) o).size() == 0;
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.AbstractMap#put(java.lang.Object, java.lang.Object)
+        /**
+         * {@inheritDoc}
          */
         @Override
         public List<? extends DNSEntry> put(String key, List<? extends DNSEntry> value) {
             return null;
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.AbstractMap#hashCode()
+        /**
+         * {@inheritDoc}
          */
         @Override
         public int hashCode() {
@@ -173,9 +150,9 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
      */
     protected static class _CacheEntry extends Object implements Map.Entry<String, List<? extends DNSEntry>> {
 
-        private List<? extends DNSEntry>  _value;
+        private List<? extends DNSEntry> _value;
 
-        private String                      _key;
+        private String                   _key;
 
         /**
          * @param key
@@ -198,64 +175,73 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
             }
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.Map$Entry#getKey()
+        /**
+         * {@inheritDoc}
          */
+        @Override
         public String getKey() {
             return (_key != null ? _key : "");
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.Map$Entry#getValue()
+        /**
+         * {@inheritDoc}
          */
+        @Override
         public List<? extends DNSEntry> getValue() {
             return _value;
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.util.Map$Entry#setValue(java.lang.Object)
+        /**
+         * {@inheritDoc}
          */
+        @Override
         public List<? extends DNSEntry> setValue(List<? extends DNSEntry> value) {
             List<? extends DNSEntry> oldValue = _value;
             _value = value;
             return oldValue;
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.lang.Object#equals(java.lang.Object)
+        /**
+         * Returns <tt>true</tt> if this list contains no elements.
+         * 
+         * @return <tt>true</tt> if this list contains no elements
          */
-        @SuppressWarnings("unchecked")
-        @Override
-        public boolean equals(Object entry) {
-            if (!(entry instanceof Map.Entry))
-                return false;
-            return this.getKey().equals(((Map.Entry) entry).getKey()) && this.getValue().equals(((Map.Entry) entry).getValue());
+        public boolean isEmpty() {
+            return this.getValue().isEmpty();
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.lang.Object#hashCode()
+        /**
+         * {@inheritDoc}
+         */
+        @Override
+        public boolean equals(Object entry) {
+            if (!(entry instanceof Map.Entry)) {
+                return false;
+            }
+            return this.getKey().equals(((Map.Entry<?, ?>) entry).getKey()) && this.getValue().equals(((Map.Entry<?, ?>) entry).getValue());
+        }
+
+        /**
+         * {@inheritDoc}
          */
         @Override
         public int hashCode() {
-            return ((_key == null) ? 0 : _key.hashCode()) ^ ((_value == null) ? 0 : _value.hashCode());
+            return (_key == null ? 0 : _key.hashCode());
         }
 
-        /*
-         * (non-Javadoc)
-         * @see java.lang.Object#toString()
+        /**
+         * {@inheritDoc}
          */
         @Override
         public synchronized String toString() {
-            StringBuffer aLog = new StringBuffer();
-            aLog.append("\n\t\tname '" + _key + "'");
-            if (_value != null) {
+            StringBuffer aLog = new StringBuffer(200);
+            aLog.append("\n\t\tname '");
+            aLog.append(_key);
+            aLog.append("' ");
+            if ((_value != null) && (!_value.isEmpty())) {
                 for (DNSEntry entry : _value) {
-                    aLog.append("\n\t\t\t" + entry.toString());
+                    aLog.append("\n\t\t\t");
+                    aLog.append(entry.toString());
                 }
             } else {
                 aLog.append(" no entries");
@@ -268,24 +254,26 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
      *
      */
     public DNSCache() {
-        super();
+        this(1024);
     }
 
     /**
      * @param map
      */
     public DNSCache(DNSCache map) {
-        this();
-        this.putAll(map);
+        this(map != null ? map.size() : 1024);
+        if (map != null) {
+            this.putAll(map);
+        }
     }
 
     /**
      * Create a table with a given initial size.
-     *
+     * 
      * @param initialCapacity
      */
     public DNSCache(int initialCapacity) {
-        this();
+        super();
         _entrySet = new HashSet<Map.Entry<String, List<? extends DNSEntry>>>(initialCapacity);
     }
 
@@ -298,8 +286,9 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
      */
     @Override
     public Set<Map.Entry<String, List<? extends DNSEntry>>> entrySet() {
-        if (_entrySet == null)
+        if (_entrySet == null) {
             _entrySet = new HashSet<Map.Entry<String, List<? extends DNSEntry>>>();
+        }
         return _entrySet;
     }
 
@@ -311,19 +300,20 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
         String stringKey = (key != null ? key.trim().toLowerCase() : null);
         for (Map.Entry<String, List<? extends DNSEntry>> entry : this.entrySet()) {
             if (stringKey != null) {
-                if (stringKey.equals(entry.getKey()))
+                if (stringKey.equals(entry.getKey())) {
                     return entry;
+                }
             } else {
-                if (entry.getKey() == null)
+                if (entry.getKey() == null) {
                     return entry;
+                }
             }
         }
         return null;
     }
 
-    /*
-     * (non-Javadoc)
-     * @see java.util.AbstractMap#put(java.lang.Object, java.lang.Object)
+    /**
+     * {@inheritDoc}
      */
     @Override
     public List<? extends DNSEntry> put(String key, List<? extends DNSEntry> value) {
@@ -339,9 +329,8 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
         }
     }
 
-    /*
-     * (non-Javadoc)
-     * @see java.util.AbstractMap#clone()
+    /**
+     * {@inheritDoc}
      */
     @Override
     protected Object clone() throws CloneNotSupportedException {
@@ -352,21 +341,22 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
 
     /**
      * Returns all entries in the cache
-     *
+     * 
      * @return all entries in the cache
      */
     public synchronized Collection<DNSEntry> allValues() {
         List<DNSEntry> allValues = new ArrayList<DNSEntry>();
         for (List<? extends DNSEntry> entry : this.values()) {
-            if (entry != null)
+            if (entry != null) {
                 allValues.addAll(entry);
+            }
         }
         return allValues;
     }
 
     /**
      * Iterate only over items with matching name. Returns an list of DNSEntry or null. To retrieve all entries, one must iterate over this linked list.
-     *
+     * 
      * @param name
      * @return list of DNSEntries
      */
@@ -375,8 +365,8 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
     }
 
     /**
-     * Get a matching DNS entry from the table (using equals). Returns the entry that was found.
-     *
+     * Get a matching DNS entry from the table (using isSameEntry). Returns the entry that was found.
+     * 
      * @param dnsEntry
      * @return DNSEntry
      */
@@ -386,7 +376,7 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
             Collection<? extends DNSEntry> entryList = this.getDNSEntryList(dnsEntry.getKey());
             if (entryList != null) {
                 for (DNSEntry testDNSEntry : entryList) {
-                    if (dnsEntry.equals(testDNSEntry)) {
+                    if (testDNSEntry.isSameEntry(dnsEntry)) {
                         result = testDNSEntry;
                         break;
                     }
@@ -398,7 +388,7 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
 
     /**
      * Get a matching DNS entry from the table.
-     *
+     * 
      * @param name
      * @param type
      * @param recordClass
@@ -409,7 +399,7 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
         Collection<? extends DNSEntry> entryList = this.getDNSEntryList(name);
         if (entryList != null) {
             for (DNSEntry testDNSEntry : entryList) {
-                if (testDNSEntry.getRecordType().equals(type) && testDNSEntry.getRecordClass().equals(recordClass)) {
+                if (testDNSEntry.getRecordType().equals(type) && ((DNSRecordClass.CLASS_ANY == recordClass) || testDNSEntry.getRecordClass().equals(recordClass))) {
                     result = testDNSEntry;
                     break;
                 }
@@ -418,9 +408,33 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
         return result;
     }
 
+    /**
+     * Get all matching DNS entries from the table.
+     * 
+     * @param name
+     * @param type
+     * @param recordClass
+     * @return list of entries
+     */
+    public synchronized Collection<? extends DNSEntry> getDNSEntryList(String name, DNSRecordType type, DNSRecordClass recordClass) {
+        Collection<? extends DNSEntry> entryList = this.getDNSEntryList(name);
+        if (entryList != null) {
+            entryList = new ArrayList<DNSEntry>(entryList);
+            for (Iterator<? extends DNSEntry> i = entryList.iterator(); i.hasNext();) {
+                DNSEntry testDNSEntry = i.next();
+                if (!testDNSEntry.getRecordType().equals(type) || ((DNSRecordClass.CLASS_ANY != recordClass) && !testDNSEntry.getRecordClass().equals(recordClass))) {
+                    i.remove();
+                }
+            }
+        } else {
+            entryList = Collections.emptyList();
+        }
+        return entryList;
+    }
+
     /**
      * Adds an entry to the table.
-     *
+     * 
      * @param dnsEntry
      * @return true if the entry was added
      */
@@ -450,38 +464,70 @@ public class DNSCache extends AbstractMap<String, List<? extends DNSEntry>>
 
     /**
      * Removes a specific entry from the table. Returns true if the entry was found.
-     *
+     * 
      * @param dnsEntry
      * @return true if the entry was removed
      */
     public synchronized boolean removeDNSEntry(DNSEntry dnsEntry) {
         boolean result = false;
         if (dnsEntry != null) {
-            Collection<? extends DNSEntry> entry = this.getDNSEntryList(dnsEntry.getKey());
-            if (entry != null) {
-                result = entry.remove(dnsEntry);
-                // If we just removed the last one we need to get rid of the
-                // entry
-                if (entry.size() == 0)
-                    this.remove(dnsEntry.getKey());
+            Map.Entry<String, List<? extends DNSEntry>> existingEntry = this.getEntry(dnsEntry.getKey());
+            if (existingEntry != null) {
+                result = existingEntry.getValue().remove(dnsEntry);
+                // If we just removed the last one we need to get rid of the entry
+                if (existingEntry.getValue().isEmpty()) {
+                    this.entrySet().remove(existingEntry);
+                }
             }
         }
         return result;
     }
 
-    /*
-     * (non-Javadoc)
-     * @see java.util.AbstractMap#toString()
+    /**
+     * Replace an existing entry by a new one.<br/>
+     * <b>Note:</b> the 2 entries must have the same key.
+     * 
+     * @param newDNSEntry
+     * @param existingDNSEntry
+     * @return <code>true</code> if the entry has been replace, <code>false</code> otherwise.
+     */
+    public synchronized boolean replaceDNSEntry(DNSEntry newDNSEntry, DNSEntry existingDNSEntry) {
+        boolean result = false;
+        if ((newDNSEntry != null) && (existingDNSEntry != null) && (newDNSEntry.getKey().equals(existingDNSEntry.getKey()))) {
+            Map.Entry<String, List<? extends DNSEntry>> oldEntry = this.getEntry(newDNSEntry.getKey());
+
+            List<DNSEntry> aNewValue = null;
+            if (oldEntry != null) {
+                aNewValue = new ArrayList<DNSEntry>(oldEntry.getValue());
+            } else {
+                aNewValue = new ArrayList<DNSEntry>();
+            }
+            aNewValue.remove(existingDNSEntry);
+            aNewValue.add(newDNSEntry);
+
+            if (oldEntry != null) {
+                oldEntry.setValue(aNewValue);
+            } else {
+                this.entrySet().add(new _CacheEntry(newDNSEntry.getKey(), aNewValue));
+            }
+            // This is probably not very informative
+            result = true;
+        }
+        return result;
+    }
+
+    /**
+     * {@inheritDoc}
      */
     @Override
     public synchronized String toString() {
-        StringBuffer aLog = new StringBuffer();
+        StringBuffer aLog = new StringBuffer(2000);
         aLog.append("\t---- cache ----");
         for (Map.Entry<String, List<? extends DNSEntry>> entry : this.entrySet()) {
-            aLog.append("\n\t\t" + entry.toString());
+            aLog.append("\n\t\t");
+            aLog.append(entry.toString());
         }
         return aLog.toString();
     }
 
-
 }
